/* Name:  knn.c

Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <time.h>
#include "knn.h"

// Output directory
char output_directory[MAX_LINE_CHARS] = ".";

// Global variables. Parsed from the configuration file
int g_K = 0; // K for k-NN
int g_feature_sel[MAX_NUM_FEATURES]; // feature selection bit vector
int g_scales[MAX_NUM_FEATURES]; // feature scaling
int g_distance_metric = 0; // distance metric: 0 - Euclidean; 1 - Hamming; 2 - Manhattan
//int g_napps = 0; // # of applications to be identified

int nfiles[2]; // # of training files &  # of test files
struct tagfile * train_list = NULL; // list of the training file names
struct tagfile * test_list = NULL; // list of the test file names

int napps = 0; // # of applications
struct tagapp * app_list = NULL; // list of applications

int nhashs = 0; // # of hash tables
struct taghash * hash_tables = NULL; // list of hash tables

// Parse the configuration file
int parse_conf(FILE *fp){
  char line[MAX_LINE_CHARS];
  char *pch, *pchb;
  int fi = MAX_NUM_FEATURES; // the index of the current feature
  int ok = 1;
  while(!feof(fp)){// parse the file
    fgets(line, sizeof(line), fp);
    pch = strchr(line, '#');
    if(pch != NULL){
      *pch = '\0';
    }
    //puts(line);
    pch = strtok(line, " :=");
    int parameter_type = 0;// the type of the parameter in the current line
    while(pch != NULL){// parse the line
      //puts(pch);
      switch(parameter_type){
	//case 1: g_napps = atoi(pch); break;
	case 2: g_K = atoi(pch); break;
	case 3: g_feature_sel[fi++] = atoi(pch); break;
	case 4: g_scales[fi++] = atoi(pch); break;
	case 5: g_distance_metric = trans_metric(pch); //puts(pch); printf("distance_metric=%d\n", g_distance_metric);
		break;
	case 6:	sscanf(pch, "%s", train_list[nfiles[0]-1].filename); break;
	case 7: sscanf(pch, "%s", test_list[nfiles[1]-1].filename); break;
        //case 8: nhashs = atoi(pch); break;
        case 9: hash_tables[nhashs-1].masks[fi++] = strtoul(pch, NULL, 0); break;
	default: break;
      }
      if(fi >= MAX_NUM_FEATURES){ //ignore the remaining content in the current line
        parameter_type = 0;
      }
      if(strstr(pch, "nApp")){
	parameter_type = 1;
      } else if(!strcmp(pch, "K")){
	parameter_type = 2;
      } else if(strstr(pch, "Features")){
	parameter_type = 3;
	fi = 0;
      } else if(strstr(pch, "Scales")){
	parameter_type = 4;
	fi = 0;
      } else if(!strcmp(pch, "Metric")){
	parameter_type = 5;
      } else if(pchb = strstr(pch, "Train")){
	parameter_type = 6;
	pchb += strlen("Train");
	//puts(pchb);
	int atype = atoi(pchb);
        nfiles[0] ++;
        train_list = (struct tagfile *)realloc(train_list, nfiles[0]*sizeof(struct tagfile));
	train_list[nfiles[0]-1].appid = insert_app_list(atype, 0);
      } else if(pchb = strstr(pch, "Test")){
	parameter_type = 7;
	pchb += strlen("Test");
	//puts(pchb);
	int atype = atoi(pchb);
        nfiles[1] ++;
        test_list = (struct tagfile *)realloc(test_list, nfiles[1]*sizeof(struct tagfile));
	test_list[nfiles[1]-1].appid = insert_app_list(atype, 1);
      } else if(strstr(pch, "nHash")){
        parameter_type = 8;
      } else if(strstr(pch, "HashBits")){
        parameter_type = 9;
        fi = 0;
    	nhashs ++;
	hash_tables = (struct taghash *)realloc(hash_tables, nhashs*sizeof(struct taghash));
      }

      pch = strtok(NULL, " :=");
    }
  }
  if(napps <= 0){
    printf("# of applications to be identified must be >= 1.\n");
    ok = 0;
  }
  if(g_K <= 0){
    ok = 0;
  }
  if(g_distance_metric < 0){
    ok = 0;
  }
  if(nfiles[0] <= 0 && nfiles[1] <= 0){// indiate neither training nor test given
    printf("Neither training nor test files are given.\n");
    ok = 0;
  }

  return ok;
}

// Insert a file with labeled application into the app list
// Return the id of the application
int insert_app_list(int atype, int ftype){
  int i;
  for(i = 0; i < napps; i ++){
    if(app_list[i].apptype == atype){
      app_list[i].nfs[ftype] ++;
      break;
    }
  }
  if(i >= napps){
    napps ++;
    app_list = (struct tagapp *)realloc(app_list, napps*sizeof(struct tagapp));
    app_list[napps-1].apptype = atype;
    app_list[napps-1].nfs[ftype] = 1;
    app_list[napps-1].nfs[1-ftype] = 0;
    app_list[napps-1].appid = napps-1;
    memset(app_list[napps-1].features_count, 0, MAX_NUM_FEATURES*(1<<MAX_FEATURE_WIDTH)*sizeof(uint));
    app_list[napps-1].nsamples = 0;
  }
  return app_list[i].appid;
}

// Initialize hash tables
int init_hash_tables(){
  printf("nHash = %d\n", nhashs);
  if(nhashs <= 0){
    printf("LSH is disabled.\n");
    return 0;
  }
  int i, j;
  for(i = 0; i < nhashs; i ++){
	printf("Hashing bits of %dth table: ", i);
	for(j = 0; j < MAX_NUM_FEATURES; j ++){
	  printf("%x\t", hash_tables[i].masks[j]);
	}
	printf("\n");
    uint hkl = sum_nbits(hash_tables[i].masks, MAX_NUM_FEATURES); // hash key length
    if(hkl <= 0){
      printf("Invalid configuration for %dth hashing table.\n", i);
      return 0;
    }
    uint hts = 1 << hkl; // hash table size;
    hash_tables[i].entries = (struct tagbucket *)malloc(hts*sizeof(struct tagbucket));
    memset(hash_tables[i].entries, 0, hts*sizeof(struct tagbucket));
    printf("Size of %dth hash table: %d\n", i, hts);
  }
  return 1;
}

// Sum up an integer array
uint sum_int(uint *a, int l){
  int i;
  uint rslt;
  for(i = 0, rslt = 0; i < l; i ++){
    if(g_feature_sel[i]){
      rslt += a[i];
    }
  }
  return rslt;
}

// number of 1's (pop count)
uint pop_count(uint a){
  int n1 = 0;
  while(a != 0){
	n1 += (a & 1);
	a = a >> 1;
  }
  return n1;
}

// Sum up the number of 1-bits
uint sum_nbits(uint *a, int l){
  int i;
  uint rslt;
  for(i = 0, rslt = 0; i < l; i ++){
    if(g_feature_sel[i]){
      rslt += pop_count(a[i]);
    }
  }
  return rslt;
}

// Value after bitwise Masking
uint vam(uint a, uint mask){
  uint v = 0;
  uint n = 0;
  while(mask != 0){
    if(mask & 1){
	  v |= ((a & 1)<<n);
	  n ++;
	}
	a = a >> 1;
    mask = mask >> 1;
  }
  return v;
}

// Generate the hash address
uint hash_key(uint *r, uint *b){
  uint k = 0;
  uint hkl = sum_nbits(b, MAX_NUM_FEATURES);
  int i;
  for(i = 0; i < MAX_NUM_FEATURES; i ++){
    if(g_feature_sel[i]){
      hkl -= pop_count(b[i]);
      k |= (vam(r[i], b[i])<<hkl);
    }
  }
  return k;
}


// See if the ip.proto of the record matches the hash table type
// return 1 if same
int ip_proto_match(uint ipp, uint *b){
  if(b[8]||b[9]||b[10]){// UDP
    if(ipp == IP_PROTO_TCP) return 0;
  } else if(b[2]||b[3]||b[4]||b[5]){// TCP
    if(ipp == IP_PROTO_UDP) return 0;
  }
  return 1;
}

// Training
int training(int bstat){
  clock_t start, end;
  time_t startt, endt;
  double elapsed;
  time(&startt);
  start = clock();
  train_files(nfiles[0], train_list, bstat);
  end = clock();
  time(&endt);
  elapsed = ((double)(end - start)) / CLOCKS_PER_SEC;
  printf("Training takes %.6lf seconds. (%.2lf seconds)\n", elapsed, difftime(endt, startt));
  return napps;
}

// Train the list of file names
// Collect statistics on each features
// Map onto hash tables
void train_files(int nf, struct tagfile * fl, int bstat){
  int i, j, k;
  for(i = 0; i < nf; i ++){
    printf("Training %s...\n", fl[i].filename);
    FILE *fp_train;
    fp_train = fopen(fl[i].filename, "r");
    if(fp_train == NULL){
      continue;
    }
    while(!feof(fp_train)){
      uint s[MAX_NUM_FEATURES];
      if(!read_record(fp_train, s)){// read one record from the file
        continue;
      }
      if(bstat){      // statistics on each feature
      statistics_features(fl[i].appid, s);
      app_list[fl[i].appid].nsamples ++;
      }

      // update hash tables
      for(j = 0; j < nhashs; j ++){
        if(!ip_proto_match(s[0], hash_tables[j].masks)) continue;
        uint addr = hash_key(s, hash_tables[j].masks); // addr to access hash table
	//	printf("addr = %d\n", addr);
        hash_tables[j].entries[addr].appid = fl[i].appid;
        for(k = 0; k < MAX_NUM_FEATURES; k ++){
          hash_tables[j].entries[addr].features[k] = s[k];
        }
      }
    }
    fclose(fp_train);
  }
}

// Read one record from the training/test file
int read_record(FILE *fp, uint *r){
  memset(r, 0, MAX_NUM_FEATURES*sizeof(uint));
  char line[MAX_LINE_CHARS];
  //char *pch;
  uint ip_proto = 0;
  if(fgets(line, sizeof(line), fp) == NULL) return 0;
  //puts(line);
  sscanf(line, "%x ", &ip_proto);
  //printf("ip.proto = %d\n", ip_proto);

  /* //read 10-tuple format files
  uint tmpid, tmplen;
  sscanf(line, "%u %u %x %x %x %u %u %u %x %u\n", r+7, r+1, &tmpid, r+5, r, r+2, r+3, r+6, r+4, &tmplen);
  */

  if(ip_proto == IP_PROTO_TCP){// TCP
    sscanf(line, "%x %u %u %u %x %u %x %u\n", r, r+1, r+2, r+3, r+4, r+5, r+6, r+7);
  } else if(ip_proto == IP_PROTO_UDP){// UDP
    sscanf(line, "%x %u %x %u %u %u %u\n", r, r+1, r+6, r+7, r+8, r+9, r+10);
  } else{
    //printf("Error reading line %d\n", __LINE__);
    return 0;
  }
  return 1;
}

// Update statistics on each feature
void statistics_features(int aid, uint *r){
  int i;
  for(i = 0; i < MAX_NUM_FEATURES; i ++){
    if(g_feature_sel[i]){
      app_list[aid].features_count[i][r[i]]++;
    }
  }
}

// Compute the Euclidean distance between two integers
uint eucl(uint x, uint y){
  return (x-y)*(x-y);
}

// Compute the Hamming distance
uint hamm(uint x, uint y){
  return pop_count(x ^ y);
  // uint z = x ^ y;
  // uint n = 0;
  // while(z != 0){
    // if(z & 1) n ++;
    // z = z >> 1;
  // }
  // return n;
}

// Compute the Manhattan distance
uint manh(uint x, uint y){
  if(x >= y){
    return x-y;
  } else{
    return y-x;
  }
}

// Compute distance between two vectors
uint distance(uint *s, uint *t){
  uint d = 0;
  int i;
  for(i = 0; i < MAX_NUM_FEATURES; i ++){
    if(g_feature_sel[i]){
      switch(g_distance_metric){
        case 0: d += eucl(s[i]/g_scales[i], t[i]/g_scales[i]); break;
	case 1: d += hamm(s[i]/g_scales[i], t[i]/g_scales[i]); break;
	case 2: d += manh(s[i]/g_scales[i], t[i]/g_scales[i]); break;
	default: break;
      }
    }
  }
  return d;
}

// Compute a test instance against a training file with only one application
// update the k nearest distances for the test instance
void kndistance(uint *t, struct tagfile f, struct tagdistance *dl){
  uint d;
  // Open the training file
  FILE *fp = fopen(f.filename, "r");
  if(fp== NULL){
    printf("Cannot open training file: %s\n", f.filename);
    return;
  }
  //printf("Training file: %s (app id %d, type %d)\n", f.filename, f.appid, app_list[f.appid].apptype);
  while(!feof(fp)){
    uint s[MAX_NUM_FEATURES];
    if(!read_record(fp, s)){// read one record from the file
      continue;
    }
    //printf("s[0]= %d, t[0] = %d\n", s[0], t[0]);
    if(s[0] != t[0]){// compare ip.proto to see if TCP or UDP
      //printf("ip.proto does not match. skip..\n");
      continue;
    }
    d = distance(s, t);
    insert_knd(d, f.appid, dl);
  }
  fclose(fp);
}

// Compute a test instance against a hash table
// update the k nearest distances for the test instance
void kndistance_lsh(uint *t, struct taghash h, struct tagdistance *dl){
  uint d;
  uint addr;
  if(!ip_proto_match(t[0], h.masks)){// compare ip.proto to see if TCP or UDP
    return;
  }
  addr = hash_key(t, h.masks);
  //printf("addr = %d\n", addr);
  //print_record(h.masks);
  //printf("appid = %d\n", h.entries[addr].appid);
  //print_record(h.entries[addr].features);
  //printf("distance = %d\n", d);

  if(h.entries[addr].features[0] <= 0) return; // empty bucket

  d = distance(h.entries[addr].features, t);
  insert_knd(d, h.entries[addr].appid, dl);
}

// Initialize the knd list before testing each instance
void init_knd(struct tagdistance *dl){
  int i;
  for(i = 0; i < g_K; i ++){
    dl[i].distance = -1;
	dl[i].appid = napps;
  }
}

// Insert a distance into knd list
// Return 1 if inserted
int insert_knd(uint d, int ai, struct tagdistance *dl){
  int i, j;
  for(i = 0; i < g_K; i ++){
    if(d < dl[i].distance){
      break;
    }
  }
  if(i >= g_K){
    return 0;
  }
  for(j = g_K-1; j > i; j --){
    dl[j].distance = dl[j-1].distance;
    dl[j].appid = dl[j-1].appid;
  }
  dl[i].distance = d;
  dl[i].appid = ai;
  return 1;
}

// Return the majority among the k nearest neighbors
int majority_knd(struct tagdistance *dl, int k){
  if(k == 1){// 1-NN
    return dl[0].appid;
  }
  int i;
  int * capp = (int *)malloc(napps * sizeof(int));
  memset(capp, 0, napps*sizeof(int));
  for(i = 0; i < k; i ++){
    if((dl[i].appid >= 0)&&(dl[i].appid < napps)){
      capp[dl[i].appid] ++;
	}
  }

  int cm = 0;
  int im = napps; // "unknown"
  for(i = 0; i < napps; i ++){
    if(capp[i] > cm){
      cm = capp[i];
      im = i;
    } /*else if(capp[i] == cm){
      im = majority_knd(dl, k-1);
    }*/
  }
  return im;
}

// Classification
// type = 0, brute-force
// type = 1, lsh
void classify(int type){
  int i, j;
  struct tagdistance * knd_list = (struct tagdistance *)malloc(g_K*sizeof(struct tagdistance));
  clock_t start, end;
  time_t startt, endt;
  double elapsed;
  uint ** nmatchs; // matrix of classification results
  FILE *fp_test;
  uint ntest = 0;

  // Initialize the results matrix
  nmatchs = (uint **)malloc(napps * sizeof(uint *));
  for(i = 0; i < napps; i ++){
    nmatchs[i] = (uint *)malloc((napps+1) * sizeof(uint));
    for(j = 0; j <= napps; j ++){
      nmatchs[i][j] = 0;
    }
  }
  // Classifying test sets
  printf("Start classification (%s)...\n", (type)?"LSH":"Brute-force");
  time(&startt);
  start = clock();

  for(i = 0; i < nfiles[1]; i ++){
    printf("----------------\nTesting the file: %s (Appid: %d)\n", test_list[i].filename, test_list[i].appid);
    fp_test = (FILE *)fopen(test_list[i].filename, "r");
    if(fp_test == NULL){
      printf("Cannot open the test file: %s\n", test_list[i].filename);
      continue;
    }
    while(!feof(fp_test)){
      uint t[MAX_NUM_FEATURES];
      if(!read_record(fp_test, t)){
	continue;
      }
      printf("Test instance:");
      print_record(t);
      ntest ++;
      init_knd(knd_list);
      if(type == 0){// brute-force
        for(j = 0; j < nfiles[0]; j ++){
          kndistance(t, train_list[j], knd_list);
        }
      } else{// lsh
        for(j = 0; j < nhashs; j ++){
          kndistance_lsh(t, hash_tables[j], knd_list);
        }
      }
      print_knds(g_K, knd_list);
      int result = majority_knd(knd_list, g_K);
      printf("Result: appid = %d\n", result);
      // Record the classification results
      nmatchs[test_list[i].appid][result] ++;
    }
    fclose(fp_test);
  }
  end = clock();
  time(&endt);
  elapsed = ((double)(end - start)) / CLOCKS_PER_SEC;
  printf("----------------\nClassification (%s) takes %.6lf seconds. (%.2lf seconds)\n", (type)?"lsh":"brute-force", elapsed, difftime(endt, startt));
  printf("Average time per instance: %.6f mseconds.\n", elapsed*1000/ntest);

  // Output results
  print_results(nmatchs);
}

// Print the list of applications
// Print the selected features
void print_features(int sel[], int scl[]){
  printf("Features/Scales being used: \n");
  printf(sel[0] ? "ip.proto/%d\t" : " /%d\t", scl[0]);
  printf(sel[1] ? "ip.len/%d\t" : " /%d\t", scl[1]);
  printf(sel[2] ? "tcp.srcport/%d\t" : " /%d\t", scl[2]);
  printf(sel[3] ? "tcp.dstport/%d\t" : " /%d\t", scl[3]);
  printf(sel[4] ? "tcp.flags/%d\t" : " /%d\t", scl[4]);
  printf(sel[5] ? "tcp.hdr_len/%d\t" : " /%d\t", scl[6]);
  printf(sel[6] ? "ip.flags/%d\t" : " /%d\t", scl[5]);
  printf(sel[7] ? "ip.hdr_len/%d\t" : " /%d\t", scl[7]);
  printf(sel[8] ? "udp.srcport/%d\t" : " /%d\t", scl[8]);
  printf(sel[9] ? "udp.dstport/%d\t" : " /%d\t", scl[9]);
  printf(sel[10] ? "udp.length/%d\t" : " /%d\t", scl[10]);
  printf("\n");
}

// Print the distance metric
void print_metric(int metric){
  printf("Distance metric being used: ");
  switch(metric){
    case 0: printf("Euclidean\n"); break;
    case 1: printf("Hamming\n"); break;
    case 2: printf("Manhattan\n"); break;
    default: printf("Unknown\n"); break;
  }
}

// Print the list of file names
void print_files(int nf, struct tagfile * fl){
  int i;
  printf("%d files\n", nf);
  for(i = 0; i < nf; i ++){
    printf("%s: AppType%d\n", fl[i].filename, app_list[fl[i].appid].apptype);
  }
}

// Print the list of applications
void print_apps(int na, struct tagapp *al){
  int i;
  printf("Actually %d applications:\n", na);
  printf("ID:\tAppType\t\t# of train files\t# of test files\n");
  for(i = 0; i < na; i ++){
    printf("%d: \t%d\t\t%d \t\t\t%d\n", al[i].appid, al[i].apptype, al[i].nfs[0], al[i].nfs[1]);
  }
}

// Translate the metric from string to int
int trans_metric(char *s){
  if(strchr(s, 'E')){
    return 0;
  } else if(strchr(s, 'H')){
    return 1;
  } else if(strchr(s, 'M')){
    return 2;
  } else{
    return -1;
  }
}

// Print one record
void print_record(uint *r){
  int i;
  for(i = 0; i < MAX_NUM_FEATURES; i ++){
    printf("%d ", r[i]);
  }
  printf("\n");
}

// Print knd list
void print_knds(int nd, struct tagdistance *dl){
  int i;
  for(i = 0; i < nd; i ++){
    if(dl != NULL){
      printf("min distance to app %d: %d\n", dl[i].appid, dl[i].distance);
    }
  }
}

// Output statistics of all features into files
void output_statistics(){
  int i, j, aid;
  char fn[MAX_LINE_CHARS];
  FILE *fp;
  int feature_width[MAX_NUM_FEATURES];
  for(i = 0; i < MAX_NUM_FEATURES; i ++){
    feature_width[i] = 8;
  }
  for(i = 1; i < 4; i ++){
    feature_width[i] = 16;
  }
  for(i = 8; i < 11; i ++){
    feature_width[i] = 16;
  }

  for(i = 0; i < MAX_NUM_FEATURES; i ++){
    if(!g_feature_sel[i]) continue;
    for(aid = 0; aid < napps; aid ++){
      sprintf(fn, "%s/app%d.feature%d.stat", output_directory, aid, i);
      fp = fopen(fn, "w");
      if(fp == NULL){
        printf("Cannot open the file %s to write.\n", fn);
        continue;
      }
      fprintf(fp, "# App %d Feature %d:\n# Value\tCount\tRatio\n", aid, i);
      for(j = 0; j < 1<<feature_width[i]; j ++){
        fprintf(fp, "%d\t%d\t%.6f\n", j, app_list[aid].features_count[i][j], (double)app_list[aid].features_count[i][j]/(double)app_list[aid].nsamples);
      }
      fclose(fp);
    }
  }
}

// Print results
void print_results(uint **nms){
  int i, j;
  uint *nmapp = (uint *) malloc(napps * sizeof(uint));
  uint ntotal = 0;
  uint ntrue = 0;

  for(i = 0; i < napps; i ++){
    nmapp[i] = 0;
    for(j = 0; j <= napps; j ++){
      nmapp[i] += nms[i][j];
    }
  }
  printf("Test \\ Result");
  for(j = 0; j < napps; j ++){
    printf("\t\tAppType%d", app_list[j].apptype);
  }
  printf("\t\tUnknown\n");
  for(i = 0; i < napps; i ++){
    printf("AppType%d: ", app_list[i].apptype);
    ntotal += nmapp[i];
    ntrue += nms[i][i];
    for(j = 0; j <= napps; j ++){
      printf("\t%d/%d (%.6f)", nms[i][j], nmapp[i], (double)nms[i][j]/(double)nmapp[i]);
    }
    printf("\n");
  }
  printf("Overall accuracy: %f\n", (double)ntrue/(double)ntotal);
}

// Main
int main(int argc, char *argv[]){
  int cls_type; // brute-force or lsh
  int b_output_stat;
  if(argc < 2){
    printf("usage: %s <config_file> [<output_directory>]\n", argv[0]);
    exit(1);
  }
  if(argc >= 3){
    b_output_stat = 1;
    strcpy(output_directory, argv[2]);
  } else{
    b_output_stat = 0;
  }

  // Read and parse the configuration file
  FILE *fp_conf;
  fp_conf = (FILE *)fopen(argv[1], "r");
  if(fp_conf == NULL){
    printf("Cannot open the configuration file: %s\n", argv[1]);
    exit(1);
  }

  nfiles[0] = 0; // # of training files
  nfiles[1] = 0; // # of test files
  if(parse_conf(fp_conf)){
    printf("Parse the configuration file successfully.\n");
    printf("nApp = %d\n", napps);
    printf("K = %d\n", g_K);
    print_features(g_feature_sel, g_scales);
    print_metric(g_distance_metric);
	print_apps(napps, app_list);
    fclose(fp_conf);
  } else{
    printf("Invalid configuration file: %s\n", argv[1]);
    exit(1);
  }

  // Initialize
  if(nfiles[0] <= 0){// no training given
    nfiles[0] = nfiles[1];
    train_list = test_list;
    printf("No training files. Use test files for training.\n");
  }
  if(nfiles[1] <= 0){// no test given
    nfiles[1] = nfiles[0];
    test_list = train_list;
    printf("No test files. Use training files for test.\n");
  }
  cls_type = init_hash_tables();
  if(!cls_type){
    printf("Fail to init hash tables.\n");
  }

  // Training
  printf("Training files:\n");
  print_files(nfiles[0], train_list);
  training(b_output_stat); // no training for the brute-force k-NN
  if(b_output_stat){
    printf("Statistics on each feature are output.\n");
    output_statistics();
  }

  // Classification
  printf("Test files:\n");
  print_files(nfiles[1], test_list);
  classify(cls_type);
  //classify(0); // brute-force
  //classify(1); // lsh

  return 0;
}













