#define MAX_NUM_FEATURES 11 // max number of features
#define MAX_LINE_CHARS 1024 // max number of chars per line
#define MAX_FEATURE_WIDTH 16 // max number of bits per feature
#define IP_PROTO_TCP 0x06 // IP protocol number for TCP
#define IP_PROTO_UDP 0x11 // IP protocol number for UDP

// struct for each file
struct tagfile{
  char filename[MAX_LINE_CHARS];
  int appid; // application index
};

// struct for each application
struct tagapp{
  int appid; // index of the app. incremented from 0
  int apptype; // application type
  int nfs[2]; // # of training files & # of test files
  uint features_count[MAX_NUM_FEATURES][1<<MAX_FEATURE_WIDTH]; // count of the values for each feature
  uint nsamples; // # of trained samples
};

// struct for each distance
struct tagdistance{
  uint distance;
  int appid;
};

// struct for each hash bucket
struct tagbucket{
  uint features[MAX_NUM_FEATURES];
  int appid;
};

// struct for each hash table
struct taghash{
  struct tagbucket *entries;
  uint masks[MAX_NUM_FEATURES];
};

// Insert a file with labeled application into the app list
// Return the id of the application
int insert_app_list(int, int);

// Compute the hamming distance between two integers
uint hamm(uint, uint);

// Compute the Euclidean distance between two integers
uint eucl(uint, uint);

// Compute the distance between two vectors
uint dist(uint *, uint *);

// Insert a distance into knd list
// Return 1 if inserted successfully
int insert_knd(uint, int, struct tagdistance *);

// Compute a test instance against a training file with only one application
// update the k nearest distances
void kndistance(uint *, struct tagfile, struct tagdistance *);

// Compute a test instance against a training application
// update the k nearest distances
void kndistance2(uint *, struct tagapp, struct tagdistance *);

// Read one record from the training/test file
int read_record(FILE *, uint *);

// Update statistics on each feature
void statistics_features(int, uint *);

// Output statistics on all features into files
void output_statistics();

// Print one record
void print_record(uint *);

// Print the selected features/scales
void print_features(int *, int *);

// Print the distance metric
void print_metric(int);

// Print the list of file names
void print_files(int, struct tagfile *);

// Print the list of applications
void print_apps(int, struct tagapp *);

// Translate the metric from string to int
int trans_metric(char *);

// Training
int training(int);

// Train the list of file names
void train_files(int, struct tagfile *, int);

// Print the list of knds for current test instance
void print_knds(int, struct tagdistance *);

// Print results
void print_results(uint **);

// Generate the hash address
uint hash_key(uint *, uint *);

// Sum up an integer array
uint sum_int(uint *, int);

// number of 1's (pop count)
uint pop_count(uint);

// Sum up the number of 1-bits
uint sum_nbits(uint *, int);

// Value after bitwise Masking
uint vam(uint, uint);

// Initialize hash tables
int init_hash_tables();

// See if the ip.proto of the record matches the hash table type
// return 1 if same
int ip_proto_match(uint, uint *);
