/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    15:21:18 02/25/2010 
// Design Name: 
// Module Name:    bmn 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: Bitonic Merge Network
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

module bmn(
    din,
    clk,
	 type,
    dout
    );

function integer log2;
  input integer value;
  begin
    value = value-1;
    for (log2=0; value>0; log2=log2+1)
      value = value>>1;
  end
endfunction

parameter ID_W = 2; // id attached to each entry
parameter VALUE_W = 6; // value to be compared
parameter N_ENTRY = 16; // # of entries to be sorted

parameter DATA_W = ID_W + VALUE_W; //Each entry consists of two parts: <value, id>
parameter N_STAGE = log2(N_ENTRY);

input [DATA_W*N_ENTRY-1:0] din;
input clk, type;
output [DATA_W*N_ENTRY-1:0] dout;

wire [DATA_W*N_ENTRY-1:0] intstg[N_STAGE:0];

assign intstg[0] = din;
assign dout = intstg[N_STAGE];

generate
	genvar i, j;
	for(i = 0; i < N_STAGE; i = i+1) begin: gen_stage
		for(j = 0; j < (1<<i); j = j+1) begin: gen_bmb
			bmb #(
			.ID_W(ID_W),
			.VALUE_W(VALUE_W),
			.N_ENTRY(N_ENTRY/(1<<i))
			) mb(
			.din(intstg[i][j*DATA_W*N_ENTRY/(1<<i) +: DATA_W*N_ENTRY/(1<<i)]),
			.clk(clk),
			.type(type),
			.dout(intstg[i+1][j*DATA_W*N_ENTRY/(1<<i) +: DATA_W*N_ENTRY/(1<<i)])
			);
		end
	end
endgenerate
 

endmodule


// BM block
module bmb(
    din,
    clk,
	 type,
    dout
    );

parameter ID_W = 2; // id attached to each entry
parameter VALUE_W = 6; // value to be compared
parameter N_ENTRY = 8; // # of entries to be sorted
parameter DATA_W = ID_W + VALUE_W; //Each entry consists of two parts: <value, id>

input [DATA_W*N_ENTRY-1:0] din;
input clk, type;
output [DATA_W*N_ENTRY-1:0] dout;

reg [DATA_W*N_ENTRY-1:0] dout;

wire [DATA_W-1:0] diarray [N_ENTRY-1:0];
wire [DATA_W-1:0] doarray [N_ENTRY-1:0];
wire [DATA_W*N_ENTRY-1:0] dcmp;

generate
	genvar i;
	for(i = 0; i < N_ENTRY/2; i = i+1) begin: gen_comp
		bcmp #(
		.ID_W(ID_W),
		.VALUE_W(VALUE_W)
		) cmp(
		.dia(diarray[i]),
		.dib(diarray[i+N_ENTRY/2]),
		.type(type),
		.dout({doarray[i+N_ENTRY/2],doarray[i]})
		);
	end
	
	for(i = 0; i < N_ENTRY; i = i+1) begin: array_conv
		assign diarray[i] = din[i*DATA_W +: DATA_W];
		assign dcmp[i*DATA_W +: DATA_W] = doarray[i];
	end
endgenerate

always @(posedge clk) begin
	dout <= dcmp;
end

endmodule

// Comparator
module bcmp(
    dia,
    dib,
	 type,
    dout
    );

parameter ID_W = 2; // id attached to each entry
parameter VALUE_W = 6; // value to be compared
parameter DATA_W = ID_W + VALUE_W; //Each entry consists of two parts: <value, id>

input [DATA_W-1:0] dia, dib;
input type;
output [DATA_W*2-1:0] dout;

reg [DATA_W*2-1:0] dout;

always @(dia, dib, type) begin
	if(type == 0) // descending
		dout = (dia[VALUE_W-1:0] < dib[VALUE_W-1:0]) ? {dib, dia} : {dia, dib};
	else // ascending
		dout = (dia[VALUE_W-1:0] < dib[VALUE_W-1:0]) ? {dia, dib} : {dib, dia};
end

//assign dout = (dia[VALUE_W-1:0] < dib[VALUE_W-1:0]) ? {dib, dia} : {dia, dib};

endmodule










