/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps

////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer:
//
// Create Date:   13:01:06 04/13/2010
// Design Name:   lshc
// Module Name:   C:/Documents and Settings/jiang7/My Documents/Research/Classification/ApplicationIdentificationKNN/ise/lshc/lshc_tb.v
// Project Name:  lshc
// Target Device:  
// Tool versions:  
// Description: 
//
// Verilog Test Fixture created by ISE for module: lshc
//
// Dependencies:
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
////////////////////////////////////////////////////////////////////////////////

module lshc_tb;

	integer ftrain, ftest, fresult;
	reg clk;

	reg enable;
	reg we;

	reg [7:0] ipproto;
	reg [15:0] pktsize;
	reg [15:0] srcport;
	reg [15:0] dstport;
	reg [1:0] appidi;
	
	reg [7:0] ipprotoB;
	reg [15:0] pktsizeB;
	reg [15:0] srcportB;
	reg [15:0] dstportB;

	reg [31:0] exp;

	// Outputs
	wire valid;
	wire [1:0] appido;
	wire [1:0] appidoB;

	integer status;
	integer i;

	lshc uut (
		.enable(enable),
		.ipproto(ipproto), 
		.pktsize(pktsize), 
		.srcport(srcport), 
		.dstport(dstport), 
		.ipprotoB(ipprotoB), 
		.pktsizeB(pktsizeB), 
		.srcportB(srcportB), 
		.dstportB(dstportB), 
		.clk(clk), 
		.appidi(appidi), 
		.we(we), 
		.appido(appido), 
		.appidoB(appidoB)
	);
	
	initial begin
	  clk = 0;
	  enable  = 0;
	  we = 0;

	  ipproto = 0;
	  pktsize = 0;
	  srcport = 0;
	  dstport = 0;
	  appidi = 0;
	  ipprotoB = 0;
	  pktsizeB = 0;
	  srcportB = 0;
	  dstportB = 0;
	  
	  exp = 0;
	  ftrain  = $fopen("1k.train.tb","r");
	  ftest   = $fopen("1k.test.tb","r");
	  fresult = $fopen("1k.result","w");
	end

	always # 5 clk = ~clk;

	
	initial begin
		 repeat (10) @ (posedge clk);
		 
		 // train
		 we = 1;
//		 while (!$feof(ftrain)) begin
		 for(i = 0; i < 4000; i = i+1) begin
			@ (negedge clk);
			enable = 1;
			status = $fscanf(ftrain,"%d\t%d\t%d\t%d\t%d\n",ipproto,pktsize,srcport,dstport,appidi);
			@ (negedge clk);
			enable = 0;
		 end
		 
		 // test
		 we = 0;
		 repeat (20) @ (posedge clk);
//		 while (!$feof(ftest)) begin
		 for(i = 0; i < 80000; i = i+1) begin
			@ (negedge clk);
			enable = 1;
			status = $fscanf(ftest,"%d\t%d\t%d\t%d\t%d\n",ipproto,pktsize,srcport,dstport,appidi);
			@ (negedge clk);
			enable = 0;
			$fwrite(fresult,"%d\t%d\t%d\t%d\t",ipproto,pktsize,srcport,dstport);
			$fwrite(fresult, "%d\n", appido);
		 end

		 repeat (10) @ (posedge clk);
		 $fclose(ftrain);
		 $fclose(ftest);
		 $fclose(fresult);
		 #100 $finish;
	end

	// debug
	initial begin
		$display("feof,\tstatus");
		$monitor("%d,\t%d", $feof(ftrain), status);
	end
endmodule



