/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    14:54:28 03/01/2010 
// Design Name: 
// Module Name:    majority_test 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module majority_test(
    appi,
    appo
    );

function integer log2;
  input integer value;
  begin
    value = value-1;
    for (log2=0; value>0; log2=log2+1)
      value = value>>1;
  end
endfunction

parameter APPID_W = 2;
parameter N_HASHTB = 4;
parameter K = 3;

input [APPID_W*N_HASHTB-1:0] appi;
output [APPID_W-1:0] appo;

assign appo = majority(appi, K);

// vote majority
function [APPID_W-1:0] majority;
	input [APPID_W*N_HASHTB-1:0] appi;
	input integer K;
	integer i;
	reg [APPID_W-1:0] apps [N_HASHTB-1:0];
	reg [log2(N_HASHTB)-1:0] nv [(1<<APPID_W)-1:0];
	reg [APPID_W-1:0] mapp;
	begin
		mapp = 0;
		for(i = 0; i < (1<<APPID_W) ; i = i+1) begin
			nv[i] = 0;
		end
		for(i = 0; (i < K) && (i < N_HASHTB); i = i+1) begin
			apps[i] = appi[i*APPID_W +: APPID_W];
			nv[apps[i]] = nv[apps[i]] + 1;
			if(nv[apps[i]] > nv[mapp]) begin
				mapp = apps[i];
			end
		end
		majority = mapp;
	end
endfunction

endmodule
