/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps

module ml555_top #(
	parameter CLK200FX_MULT = 5,
	parameter CLK200FX_DIV  = 8,
	parameter TEST_DEPTH_W = 8, // 256 test vectors
	parameter TEST_VECTOR_W = 116 // width of a test vector
	) (             
	output [2:0] PIN_LEDS,
               
	input        CLK200_N,
	input        CLK200_P,
	input        SYS_CLK_N,
	input        SYS_CLK_P,
	input        SYS_RESET_IN
);

   wire [2:0] w_leds;
   wire clk66, clk200, clk400, clkfx, RST;
   wire [TEST_VECTOR_W-1:0] tstv;
	wire [1:0] appo, appoB;

   //reg [TEST_DEPTH_W-1:0]  r_count_200; // synthesis attribute keep of r_count_200 is true; 
   reg [TEST_DEPTH_W-1:0]  r_count_fx;  // synthesis attribute keep of r_count_fx is true; 
   
   
   

   //Buffers for pins
   OBUF OBUF_p0( .I(w_leds[0]),   .O(PIN_LEDS[0]));
   OBUF OBUF_p1( .I(w_leds[1]),   .O(PIN_LEDS[1]));
   OBUF OBUF_p2( .I(w_leds[2]),   .O(PIN_LEDS[2]));

   
 

   //Clocks
   clocks #( 
		.CLK200FX_MULT    (CLK200FX_MULT),
      .CLK200FX_DIV     (CLK200FX_DIV)  
		) clocks0 (
		.CLK200_N     (CLK200_N),
		.CLK200_P     (CLK200_P),
		.SYS_RESET_IN (SYS_RESET_IN),
		.CLK66        (clk66),
		.CLK200       (clk200),
		.CLK400       (clk400),
		.CLKFX        (clkfx),
		.RST          (RST)
		);


   assign w_leds[1:0] = appo;
	assign w_leds[2] = & appoB;
   

/*   //Count at 200MHz
   always@(posedge clk200) begin
      if(RST) r_count_200 <= 0;
      else    r_count_200 <= r_count_200 + 1;
   end
*/

   //Count at fx speed
   always@(posedge clkfx) begin
      if(RST) r_count_fx <= 0;
      else    r_count_fx <= r_count_fx + 1;
   end
   

	// ROM
	rom #(
	.DATA_W(TEST_VECTOR_W),
	.ADDR_W(TEST_DEPTH_W)
	) testrom (
	.clk (clkfx),
	.addr(r_count_fx),
	.dout(tstv)
	);
   

	// LSH-Classifier
	lshc lshc_uut (
	.enable(tstv[0]),
	.ipproto(tstv[115:108]),
	.pktsize(tstv[107:92]),
   .srcport(tstv[91:76]),
   .dstport(tstv[75:60]),
	.ipprotoB(tstv[59:52]),
   .pktsizeB(tstv[51:36]),
   .srcportB(tstv[35:20]),
   .dstportB(tstv[19:4]),
   .clk(clkfx),
   .appidi(tstv[3:2]),
   .we(tstv[1]),
   .appido(appo),
	.appidoB(appoB)
   );
	
   
endmodule
