/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    09:55:52 02/25/2010 
// Design Name: 
// Module Name:    popcount 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module popcount(
	 din,
	 dout
    );

parameter IN_W = 16;
parameter OUT_W = log2(IN_W)+1; // e.g. din = 'hFFFF has 16 bits: 16 = 'h10 needs 5 bits

input [IN_W-1:0] din;
output [OUT_W-1:0] dout;

/*assign dout = popcount(din);
function [OUT_W-1:0] popcount;
	input [IN_W-1:0] din;
	integer i, j;
	reg [IN_W-1:0] temp;
	begin
		temp = din;
		for(i = 1; i <= log2(IN_W); i = i+1) begin
			for(j = 1; j <= IN_W/(1<<i); j = j+1) begin
				temp[j*(i+1)-1:(j-1)*(i+1)] = temp[2*j*i-1:(2*j-1)*i] + temp[(2*j-1)*i-1:(2*j-2)*i];
			end
		end
		popcount = temp;
	end
endfunction*/

wire [IN_W-1:0] temp [OUT_W-1:0];

assign temp[0] = din;

assign dout = temp[OUT_W-1][OUT_W-1:0];

generate
	genvar i, j;
	for(i = 1; i <= log2(IN_W); i = i+1) begin : gen_level
		for(j = 1; j <= IN_W/(1<<i); j = j+1) begin: gen_adder
			adder #(.IN_W(i)) la(
			.dia(temp[i-1][2*j*i-1:(2*j-1)*i]), 
			.dib(temp[i-1][(2*j-1)*i-1:(2*j-2)*i]), 
			.dout(temp[i][j*(i+1)-1:(j-1)*(i+1)])
			);
		end
	end
endgenerate

// Log2
function integer log2;
  input integer value;
  begin
    value = value-1;
    for (log2=0; value>0; log2=log2+1)
      value = value>>1;
  end
endfunction

endmodule

// Adder 
module adder(
	dia,
	dib,
	dout
	);

parameter IN_W = 1;
parameter OUT_W = IN_W+1;

input [IN_W-1:0] dia, dib;
output [OUT_W-1:0] dout;

assign dout = dia + dib;

endmodule









