/* Copyright (c) 2010 by Lawrence Livermore National Security, LLC.  
Produced at the Lawrence Livermore National Laboratory.  Written 
by Weirong Jiang.

LLNL-CODE-433968 All rights reserved.

This file is part of the Locality Sensitive Hashing -based traffic classifier (LSH-TC).

Please also read Our Notice and GNU General Public License that came with this distribution.  

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License (as published by the 
Free Software Foundation) version 2, dated June 1991.  

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the IMPLIED WARRANTY OF 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
terms and conditions of the GNU General Public License for more details.  

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    15:02:51 02/24/2010 
// Design Name: 
// Module Name:    sorting 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module sorting(
	 idi,
	 vi,
	 clk,
	 ido
    );
	 
parameter ID_W = 2; // # of applications to be classified: 2^APPID_W 
parameter VALUE_W = 6; //Sum of 3 distances each of which has log2(FEATURE_W) bits on a feature
parameter N_ENTRY = 16; // # of entries to be sorted

parameter DATA_W = ID_W + VALUE_W; //Each entry consists of two parts: <value, id>

input [ID_W*N_ENTRY-1:0] idi;
input [VALUE_W*N_ENTRY-1:0] vi;
input clk;
output [ID_W*N_ENTRY-1:0] ido;

wire [DATA_W*N_ENTRY-1:0] din;
wire [DATA_W*N_ENTRY-1:0] dout;
wire [VALUE_W*N_ENTRY-1:0] vo;

generate
	genvar i;
	for(i = 0; i < N_ENTRY; i = i+1) begin: data_conv
		assign din[i*DATA_W +: DATA_W] = {idi[i*ID_W +: ID_W], vi[i*VALUE_W +: VALUE_W]};
		assign {ido[i*ID_W +: ID_W], vo[i*VALUE_W +: VALUE_W]} = dout[i*DATA_W +: DATA_W];
	end
endgenerate

bsn #(
	.ID_W(ID_W),
	.VALUE_W(VALUE_W),
	.N_ENTRY(N_ENTRY)
	) sn (
	.din(din),
	.clk(clk),
	.dout(dout)
	);

endmodule



// Bitonic sorting network
module bsn(
	din,
	clk,
	dout
	);

function integer log2;
  input integer value;
  begin
    value = value-1;
    for (log2=0; value>0; log2=log2+1)
      value = value>>1;
  end
endfunction
	
parameter ID_W = 2; // id attached to each entry
parameter VALUE_W = 6; // value to be compared
parameter N_ENTRY = 16; // # of entries to be sorted

parameter DATA_W = ID_W + VALUE_W; //Each entry consists of two parts: <value, id>
parameter N_STAGE = log2(N_ENTRY);

input [DATA_W*N_ENTRY-1:0] din;
input clk;
output [DATA_W*N_ENTRY-1:0] dout;

wire [DATA_W*N_ENTRY-1:0] intstg[N_STAGE:0];
wire types[1:0];

assign types[0] = 1'b0;
assign types[1] = 1'b1;

assign intstg[0] = din;
assign dout = intstg[N_STAGE];

generate
	genvar i, j;
	for(i = 0; i < N_STAGE; i = i+1) begin: gen_stage
		for(j = 0; j < N_ENTRY/(2<<i); j = j+1) begin: gen_bmn
			bmn #(
			.ID_W(ID_W),
			.VALUE_W(VALUE_W),
			.N_ENTRY(2<<i)
			) mn(
			.din(intstg[i][j*DATA_W*(2<<i) +: DATA_W*(2<<i)]),
			.clk(clk),
			.type(types[j%2]),
			.dout(intstg[i+1][j*DATA_W*(2<<i) +: DATA_W*(2<<i)])
			);
		end
	end
endgenerate
 
endmodule










